angular.module('beamng.apps').directive('flightInstruments', [function () {
	var settingsPath = "settings/ui_apps/aviator331_flightInstruments/settings.json";

    // Data extraction
    // =============================================

    var luaCode = `(function ()
        local airVelocity = obj:getVelocity() - obj:getFlow()
        local roll, pitch, yaw = obj:getRollPitchYaw()
        return { 
            -- based on official IndicatedAirspeed
            trueAirspeed = math.sqrt(airVelocity:squaredLength() * obj:getDensityAtNode(0) / 1.225),
            verticalSpeed = airVelocity.z,
            isUpsideDown = obj:getDirectionVectorUp().z <= 0,
            yaw = yaw,
            pitch = pitch,
            roll = roll
        }
    end)()`;

    function wrapAngleInDegrees(angle) {
        return angle < 0 ? angle + 360 : angle;
    }

    function normalizeAngleDegrees(angle) {
        if (angle > 180) return angle - 360;
        if (angle < -180) return angle + 360;
        return angle;
    }

    function getSensorData(sensors) {
        return {
            // indicated airspeed is computed elsewhere
            
            // pitch as degrees, + points nose up, - points nose down
            attitudePitch: Math.asin(sensors.pitch) * 180 / Math.PI,
            // bank as degrees, + is clockwise
            attitudeBankRaw: -Math.asin(sensors.roll) * 180 / Math.PI,

            // altitude in meters, positive above sea level
            altitude: sensors.position.z,

            // heading as degrees clockwise from north = 0
            heading: wrapAngleInDegrees(-sensors.yaw * 180 / Math.PI),

            // vertical speed is computed elsewhere

            // gravity
            gravity: -sensors.gravity
        };
    }

    function getElectricsData(electrics) {
        return {
            // airspeed
            airspeed: electrics.airflowspeed,

            // altitude in meters, positive above sea level
            altitude: electrics.altitude,
        };
    }

    // Data processing
    // =============================================

    class RunningAverage {
        constructor(length) {
            this.length = length;
            this._data = new Array(length).fill(0);
            this._index = 0;
            this._sum = 0;
        }

        average() {
            return this._sum / this.length;
        }

        update(value) {
            this._sum += value - this._data[this._index];
            this._data[this._index] = value;
            this._index = (this._index + 1) % this.length;
            return this.average();
        }
    }

    // Layout
    // =============================================

    function makeInstrumentPanel(kind) {
        return `
            <div id="${kind}-parent" class="panelParent">
                <div class="panelAnalog">
                    <object id="${kind}-svg" type="image/svg+xml" data="/ui/modules/apps/FlightInstruments/${kind}.svg"></object>
                </div>
                <div class="panelDigital" id="${kind}-digital">---</div>
            </div>
        `;
    }

    // Instrument code
    // =============================================

    var svgCache = {};

    var INSTRUMENTS = ['airspeed', 'attitudeIndicator', 'altimeter', 'turnCoordinator', 'headingIndicator', 'variometer'];

    /* const */ var KNOTS_IN_METERS_PER_SECOND = 1.94384449;
    /* const */ var FEET_PER_MINUTE_IN_METERS_PER_SECOND = 196.850394;
    /* const */ var FEET_IN_METERS = 3.2808399;

    function resolveById(svg, id) {
        var cache = svgCache[svg];
        var result = cache && cache[id];

        if (cache && !result) {
            result = svg.getElementById(id);
            cache[id] = result;
        }

        return result;
    }

    function updateAirspeed(data, svgList, digital) {
        var kIAS = data.indicatedAirspeed * KNOTS_IN_METERS_PER_SECOND;

        for (var svg of svgList) {
            var pointer = resolveById(svg, 'pointer');

            if (pointer) {
                var angle;
                var TO40 = 14;
                var PAST40 = 36;
                if (kIAS < 40) {
                    angle = kIAS / 40;
                    angle *= angle * TO40;
                } else {
                    angle = TO40 + PAST40 * ((kIAS - 40) / 20);
                }
                pointer.style.transform = `rotate(${angle}deg)`;
            }
        }

        if (digital) digital.textContent = `${kIAS.toFixed(2)} kt`;
    }

    function updateAltimeter(data, svgList, digital) {
        var ftA = data.altitude * FEET_IN_METERS;

        for (var svg of svgList) {
            var ft100 = resolveById(svg, 'ft100');
            if (ft100) ft100.style.transform = `rotate(${ftA * 0.36}deg)`;

            var ft1k = resolveById(svg, 'ft1k');
            if (ft1k) ft1k.style.transform = `rotate(${ftA * 0.036}deg)`;

            var ft10k = resolveById(svg, 'ft10k');
            if (ft10k) ft10k.style.transform = `rotate(${ftA * 0.0036}deg)`;
        }

        if (digital) digital.textContent = `${ftA.toFixed(1)} ft`;
    }

    function updateAttitudeIndicator(data, svgList, digital) {
        for (var svg of svgList) {
            var NOTCH_HEIGHT = 20;
            var horizonTranslate = data.attitudePitch * NOTCH_HEIGHT / 10;

            var horizon = resolveById(svg, 'horizon');
            if (horizon) horizon.style.transform = `translate(0px, ${horizonTranslate}px)`;

            var bank = resolveById(svg, 'bank');
            if (bank) bank.style.transform = `translate(0px, ${-horizonTranslate}px) rotate(${-data.attitudeBank}deg) translate(0px, ${horizonTranslate}px)`;

            var wheel = resolveById(svg, 'wheel');
            if (wheel) wheel.style.transform = `rotate(${-data.attitudeBank}deg)`;            
        }

        if (digital) digital.textContent = `Pitch: ${data.attitudePitch.toFixed(2)} \u00b0`;
    }

    var turnCoordinatorSlipAngle = new RunningAverage(2);

    function updateTurnCoordinator(data, svgList, digital) {
        for (var svg of svgList) {
            var rateOfTurn = data.yawAngularVelocity;

            var plane = resolveById(svg, 'plane');
            if (plane) {
                var rollAngle = data.attitudeBank;
                var turnAngle = (rateOfTurn * 20 / 3 * 0.634 + rollAngle * 0.366);
                plane.style.transform = `rotate(${turnAngle}deg)`;
            }

            var slip = resolveById(svg, 'slip');
            if (slip) {
                var tanBankAngle = Math.tan(data.attitudeBankRaw * Math.PI / 180);
                var rotConstant = 180 * data.gravity / Math.PI;
                var expectedRateOfTurn = (rotConstant * tanBankAngle) / Math.max(1, data.trueAirspeed);
                var slipAngle = (rateOfTurn - expectedRateOfTurn) * 1.5;
                slipAngle = Math.min(15, Math.max(-15, turnCoordinatorSlipAngle.update(slipAngle)));
                slip.style.transform = `rotate(${slipAngle}deg)`;
            }
        }

        if (digital) digital.textContent = `Bank: ${data.attitudeBank.toFixed(2)} \u00b0`;
    }

    function updateHeadingIndicator(data, svgList, digital) {
        for (var svg of svgList) {
            var circle = resolveById(svg, 'circle');
            if (circle) circle.style.transform = `rotate(${-data.heading}deg)`;
        }

        if (digital) digital.textContent = `${data.heading.toFixed(2)} \u00b0`;
    }

    function updateVariometer(data, svgList, digital) {
        var vasi = data.verticalSpeed * FEET_PER_MINUTE_IN_METERS_PER_SECOND;

        for (var svg of svgList) {
            var pointer = resolveById(svg, 'vasi');
            if (pointer) pointer.style.transform = `rotate(${vasi * 0.072}deg)`;
        }

        if (digital) digital.textContent = `${vasi.toFixed(1)} ft/min`;
    }
    
    // table
    var html =
        `<div style="width:100%; height:100%; box-sizing:border-box;" class="bngApp aviator331-FlightInstruments">
            <link type="text/css" rel="stylesheet" href="/ui/modules/apps/FlightInstruments/app.css">
            <div class="mainPanel">
                ${INSTRUMENTS.map(instrument => makeInstrumentPanel(instrument)).join('')}
            </div>
            <div class="buttons">
                <md-button class="md-raised button-alt" ng-click="setSeaLevel()">Set sea level</md-button>
            </div>
        </div>`;

    // Entry point
    // =============================================

    // common variables        
    var seaLevel = 0;

    return {
        template: html,
        replace: true,
        restrict: 'EA',
        link: function (scope, element, attrs) {
            var streamsList = ['sensors', 'electrics'];

            StreamsManager.add(streamsList);

            var parent = null;

            var svgs = {};
            var digitals = {};

            var trueAirspeed = 0;      // m/s
            var verticalSpeed = 0;     // m/s
            var lastAltitude = 0;
            var isUpsideDown = false;
            var yawAngularVelocity = 0;    // rad/s?
            var rollAngularVelocity = 0;    // rad/s?

            var smoothedVerticalSpeed = new RunningAverage(5);
            var smoothedYawAngularVelocity = new RunningAverage(4);
            var smoothedRollAngularVelocity = new RunningAverage(10);
            
            /* bngApi.engineLua('jsonReadFile(' + bngApi.serializeToLua('/settings/ui_apps/Fuel Flow.json') + ')', (settings) => {
                appSettings = settings;
            }); */
            
            scope.setSeaLevel = function () {
                seaLevel = lastAltitude;
            }

            function addSvgForInstrument(instrument, svg) {
                svgCache[svg] = {};
                svgs[instrument].push(svg);
            }

            function discoverSvg(instrument) {
                var svg = parent.querySelector(`#${instrument}-svg`);
                if (svg) {
                    angular.element(svg).on('load', function() {
                        addSvgForInstrument(instrument, svg.contentDocument);
                    });

                    if (svg.contentDocument) {
                        addSvgForInstrument(instrument, svg.contentDocument);
                    }
                } else {
                    // element not truly ready yet
                    setTimeout(function () { discoverSvg(instrument); }, 100);
                }
            }

            element.ready(function () {
                parent = element[0];

                for (var instrument of INSTRUMENTS) {
                    svgs[instrument] = [];
                    digitals[instrument] = parent.querySelector(`#${instrument}-digital`);
                    discoverSvg(instrument);
                }
            });

            scope.$on('$destroy', function () {
                // bngApi.engineLua('jsonWriteFile(' + bngApi.serializeToLua(settingsPath) + ', ' + bngApi.serializeToLua(appSettings) + ', ' + bngApi.serializeToLua(true) + ')');
                StreamsManager.remove(streamsList);
            });

            var lastNow = null;
            var lastYaw = null, lastPitch = null, lastRoll = null;
            var timeScale = 0;

            function updateInstrument(fn, name, data) {
                fn(data, svgs[name], digitals[name]);
            }

            scope.$on('streamsUpdate', function (event, streams) {
                if (!streams.sensors || !parent) return;

                bngApi.engineLua('(be:getSimulationTimeScale())',
                    function (luaData) {
                        timeScale = luaData;
                    });

                var now = Date.now() * 1e-3;
                var delta = lastNow == null ? 0.1 : now - lastNow;
                lastNow = now;

                delta *= timeScale;

                bngApi.activeObjectLua(luaCode,
                    function (luaData) {
                        trueAirspeed = luaData.trueAirspeed;
                        verticalSpeed = luaData.verticalSpeed;
                        isUpsideDown = luaData.isUpsideDown;

                        if (delta > 0) {
                            var yaw = luaData.yaw * 180 / Math.PI;
                            // var pitch = luaData.pitch * 180 / Math.PI;
                            var roll = luaData.roll * 180 / Math.PI;

                            if (lastYaw != null) yawAngularVelocity = -normalizeAngleDegrees(yaw - lastYaw) / delta;
                            if (lastRoll != null) rollAngularVelocity = -normalizeAngleDegrees(roll - lastRoll) / delta;
                            smoothedYawAngularVelocity.update(yawAngularVelocity);
                            smoothedRollAngularVelocity.update(rollAngularVelocity);

                            lastYaw = yaw;
                            lastRoll = roll;
                        }
                    });

                var data = getSensorData(streams.sensors);

                data.trueAirspeed = trueAirspeed;

                if (streams.electrics) {
                    var edata = getElectricsData(streams.electrics);
                    if (edata.altitude != null) data.altitude = edata.altitude;
                    if (edata.airspeed != null) data.trueAirspeed = edata.airspeed;
                }

                smoothedVerticalSpeed.update(verticalSpeed);

                data.verticalSpeed = smoothedVerticalSpeed.average();
                data.yawAngularVelocity = smoothedYawAngularVelocity.average();
                data.rollAngularVelocity = smoothedRollAngularVelocity.average();

                lastAltitude = data.altitude;
                data.altitude -= seaLevel;

                // fix bank
                if (isUpsideDown) {
                    var angle = data.attitudeBankRaw;
                    angle = 180 - angle;
                    if (angle > 180) angle = angle - 360;
                    data.attitudeBank = angle;
                } else {
                    data.attitudeBank = data.attitudeBankRaw;
                }

                // simulate altitude effects on airspeed
                data.indicatedAirspeed = data.trueAirspeed / Math.max(0.5, 1 + (data.altitude * 6.56168e-05));

                updateInstrument(updateAirspeed, 'airspeed', data);
                updateInstrument(updateAltimeter, 'altimeter', data);
                updateInstrument(updateAttitudeIndicator, 'attitudeIndicator', data);
                updateInstrument(updateTurnCoordinator, 'turnCoordinator', data);
                updateInstrument(updateHeadingIndicator, 'headingIndicator', data);
                updateInstrument(updateVariometer, 'variometer', data);
            });
        }
    };
}])